;*=====================================================================*/
;*    serrano/prgm/project/bigloo/runtime/Llib/socket.scm              */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Mon Jun 29 18:45:17 1998                          */
;*    Last change :  Wed May  7 06:51:27 2003 (serrano)                */
;*    -------------------------------------------------------------    */
;*    Socket handling. This part is mostly compatible with             */
;*    STk. This code is extracted from STk by Erick Gallesio.          */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module __socket

   (import  __error)
   
   (use     __type
	    __bigloo
	    __tvector
	    __ucs2
	    __dsssl
	    
	    __r4_numbers_6_5_fixnum
	    __r4_numbers_6_5_flonum
	    __r4_equivalence_6_2
	    __r4_vectors_6_8
	    __r4_booleans_6_1
	    __r4_characters_6_6
	    __r4_symbols_6_4
	    __r4_pairs_and_lists_6_3
	    __r4_strings_6_7
	    __r4_ports_6_10_1
	    __r4_control_features_6_9

	    __evenv)
 
   (extern  (macro c-socket?::bool            (::obj)    "SOCKETP")
	    (macro c-socket-hostname::obj     (::socket) "SOCKET_HOSTNAME")
	    (macro c-socket-hostip::obj       (::socket) "SOCKET_HOSTIP")
	    (macro c-socket-down?::bool       (::socket) "SOCKET_DOWNP")
	    (macro c-socket-port-number::int  (::socket) "SOCKET_PORT")
	    (macro c-socket-input::obj        (::socket) "SOCKET_INPUT")
	    (macro c-socket-output::obj       (::socket) "SOCKET_OUTPUT")
	    
	    (c-make-client-socket::socket    (::bstring ::int ::bool)
					     "make_client_socket")
	    (c-make-server-socket::socket    (::int)
					     "make_server_socket")
	    (c-socket-accept-connection::obj (::socket ::bool)
					     "socket_accept_connection")
	    (c-socket-accept::obj            (::socket ::bool ::bool)
					     "socket_accept")
	    (c-socket-local-addr::bstring    (::socket)
					     "socket_local_addr")
	    (c-socket-dup::socket            (::socket)
					     "socket_dup")
	    (c-socket-shutdown::obj          (::socket ::bool)
					     "socket_shutdown")
	    (c-socket-close::obj             (::socket)
					     "socket_close")
	    (macro c-socket-server?::bool    (::obj) "SOCKET_SERVERP")
	    (macro c-socket-client?::bool    (::obj) "SOCKET_CLIENTP"))

   (java    (class foreign
	       (method static c-socket?::bool (::obj)
		       "SOCKETP")
	       
	       (method static c-socket-dup::socket (::socket)
		       "socket_dup")
	       (method static c-make-client-socket::socket (::bstring ::int ::bool)
		       "make_client_socket")
	       (method static c-make-server-socket::socket (::int)
		       "make_server_socket")
	       
	       (method static c-socket-hostname::obj (::socket)
		       "SOCKET_HOSTNAME")
	       (method static c-socket-hostip::obj (::socket)
		       "SOCKET_HOSTIP")
	       (method static c-socket-down?::bool (::socket)
		       "SOCKET_DOWNP")
	       (method static c-socket-port-number::int (::socket)
		       "SOCKET_PORT")
	       (method static c-socket-input::obj (::socket)
		       "SOCKET_INPUT")
	       (method static c-socket-output::obj (::socket)
		       "SOCKET_OUTPUT")
	       
	       (method static c-socket-accept-connection::obj (::socket ::bool)
		       "socket_accept_connection")
	       (method static c-socket-accept::socket (::socket ::bool ::bool)
		       "socket_accept")
	       (method static c-socket-local-addr::bstring (::socket)
		       "socket_local_addr")
	       (method static c-socket-shutdown::obj (::socket ::bool)
		       "socket_shutdown")
	       (method static c-socket-close::obj (::socket)
		       "socket_close")
	       (method static c-socket-server?::bool (::obj)
		       "SOCKET_SERVERP")
	       (method static c-socket-client?::bool (::obj)
		       "SOCKET_CLIENTP")))

   (export  (inline socket?::bool ::obj)
	    (inline socket-server?::bool ::obj)
	    (inline socket-client?::bool ::obj)
	    (inline socket-hostname::obj ::socket)
	    (inline socket-host-address::obj ::socket)
	    (inline socket-down?::bool ::socket)
	    (inline socket-port-number::bint ::socket)
	    (inline socket-input::obj ::socket)
	    (inline socket-output::obj ::socket)
	    (inline make-client-socket::socket ::bstring ::int . opt)
	    (inline make-server-socket::socket . port-number)
	    (inline socket-accept-connection ::socket . opt)
	    (inline socket-accept::obj ::socket . opt)
	    (inline socket-local-address ::socket)
	    (inline socket-dup::socket ::socket)
	    (inline socket-shutdown::obj ::socket . close)
	    (inline socket-close::obj ::socket))

   (pragma  (c-socket? nesting)
	    (c-socket-hostname nesting)
	    (c-socket-hostip nesting)
	    (c-socket-down? nesting)
	    (c-socket-port-number nesting)
	    (c-socket-input nesting)
	    (c-socket-output nesting)))

;*---------------------------------------------------------------------*/
;*    socket? ...                                                      */
;*---------------------------------------------------------------------*/
(define-inline (socket? obj)
   (c-socket? obj))

;*---------------------------------------------------------------------*/
;*    socket-server? ...                                               */
;*---------------------------------------------------------------------*/
(define-inline (socket-server? obj)
   (c-socket-server? obj))

;*---------------------------------------------------------------------*/
;*    socket-client? ...                                               */
;*---------------------------------------------------------------------*/
(define-inline (socket-client? obj)
   (c-socket-client? obj))

;*---------------------------------------------------------------------*/
;*    socket-hostname ...                                              */
;*---------------------------------------------------------------------*/
(define-inline (socket-hostname socket)
   (c-socket-hostname socket))

;*---------------------------------------------------------------------*/
;*    socket-host-address ...                                          */
;*---------------------------------------------------------------------*/
(define-inline (socket-host-address socket)
   (c-socket-hostip socket))

;*---------------------------------------------------------------------*/
;*    socket-local-address ...                                         */
;*---------------------------------------------------------------------*/
(define-inline (socket-local-address socket::socket)
   (c-socket-local-addr socket))

;*---------------------------------------------------------------------*/
;*    socket-down? ...                                                 */
;*---------------------------------------------------------------------*/
(define-inline (socket-down?::bool socket::socket)
   (c-socket-down? socket))

;*---------------------------------------------------------------------*/
;*    socket-port-number ...                                           */
;*---------------------------------------------------------------------*/
(define-inline (socket-port-number::bint socket::socket)
   (c-socket-port-number socket))

;*---------------------------------------------------------------------*/
;*    socket-input ...                                                 */
;*---------------------------------------------------------------------*/
(define-inline (socket-input socket::socket)
   (c-socket-input socket))

;*---------------------------------------------------------------------*/
;*    socket-output ...                                                */
;*---------------------------------------------------------------------*/
(define-inline (socket-output socket::socket)
   (c-socket-output socket))

;*---------------------------------------------------------------------*/
;*    make-client-socket ...                                           */
;*---------------------------------------------------------------------*/
(define-inline (make-client-socket::socket hostname port-number . buffered)
   (c-make-client-socket hostname
			 port-number
			 (or (null? buffered) (eq? (car buffered) #t))))

;*---------------------------------------------------------------------*/
;*    make-server-socket ...                                           */
;*---------------------------------------------------------------------*/
(define-inline (make-server-socket::socket . port-number)
   (c-make-server-socket (if (null? port-number) 0 (car port-number))))

;*---------------------------------------------------------------------*/
;*    socket-accept-connection ...                                     */
;*---------------------------------------------------------------------*/
(define-inline (socket-accept-connection socket::socket . buffered)
   (let ((buf (or (null? buffered) (eq? (car buffered) #t))))
      (c-socket-accept-connection socket buf)))
	    
;*---------------------------------------------------------------------*/
;*    socket-accept ...                                                */
;*---------------------------------------------------------------------*/
(define-inline (socket-accept socket::socket . buffered)
   (let ((buf (or (null? buffered) (eq? (car buffered) #t)))
	 (errp (or (null? buffered)
		   (null? (cdr buffered))
		   (cadr buffered))))
      (c-socket-accept socket buf errp)))
	    
;*---------------------------------------------------------------------*/
;*    socket-dup ...                                                   */
;*---------------------------------------------------------------------*/
(define-inline (socket-dup::socket socket::socket)
   (c-socket-dup socket))

;*---------------------------------------------------------------------*/
;*    socket-shutdown ...                                              */
;*---------------------------------------------------------------------*/
(define-inline (socket-shutdown socket::socket . close)
   (c-socket-shutdown socket (if (pair? close) (car close) #t)))

;*---------------------------------------------------------------------*/
;*    socket-close ...                                                 */
;*---------------------------------------------------------------------*/
(define-inline (socket-close socket::socket)
   (c-socket-close socket))
